/*! Tripetto Collector 1.7.1-18 - Copyright (C) 2019 Tripetto B.V. - All Rights Reserved */

/// <reference path="marblecore.hookup.d.ts" />
/// <reference path="marblecore.functional.d.ts" />

declare module "@marblecore/matrix" {
    /** Explicit exports */
    export { ICursor, ISafeCursor } from "@marblecore/matrix/lib/cursor";
    export { Matrix } from "@marblecore/matrix/lib/matrix";
    export { Item } from "@marblecore/matrix/lib/item";
    export { StrictMatrix } from "@marblecore/matrix/lib/strict/matrix";
    export { StrictItem } from "@marblecore/matrix/lib/strict/item";
    export {
        IStrictItemConstructor
    } from "@marblecore/matrix/lib/strict/constructor";
    export { Model } from "@marblecore/matrix/lib/model/model";
    export {
        ModelItem,
        IDataChangeEvent
    } from "@marblecore/matrix/lib/model/item";
    /** Hooks */
    export {
        IDimensionChangeEvent,
        ICountChangeEvent,
        IItemPushEvent,
        IItemChangeEvent,
        IItemPopEvent,
        ICursorChangeEvent
    } from "@marblecore/matrix/lib/events";
    /** Enumerations */
    export { Orientations } from "@marblecore/matrix/enum/orientations";
    export { Reasons } from "@marblecore/matrix/enum/reasons";
    export { Types } from "@marblecore/matrix/enum/types";
}

declare module "@marblecore/matrix/lib/cursor" {
    export interface ICursor {
        /** Specifies the current column of the cursor. */
        column?: number;
        /** Specifies the current row of the cursor. */
        row?: number;
        /** Specifies the current index of the cursor. */
        index?: number;
    }
    export interface ISafeCursor {
        /** Specifies the current column of the cursor. */
        column: number;
        /** Specifies the current row of the cursor. */
        row: number;
        /** Specifies the current index of the cursor. */
        index: number;
    }
}

declare module "@marblecore/matrix/lib/matrix" {
    import { HookProps, Hooks, THooks } from "@marblecore/hookup";
    import { Item } from "@marblecore/matrix/lib/item";
    import { ICursor, ISafeCursor } from "@marblecore/matrix/lib/cursor";
    import {
        ICountChangeEvent,
        ICursorChangeEvent,
        IDimensionChangeEvent,
        TMatrixHooks
    } from "@marblecore/matrix/lib/events";
    import { Orientations } from "@marblecore/matrix/enum/orientations";
    import { Types } from "@marblecore/matrix/enum/types";
    /**
     * Implements the convolution matrix.
     * @param T Specifies the item type.
     * @param H Specifies custom hooks for the matrix (optional).
     */
    export abstract class Matrix<
        T extends Item<THooks>,
        H extends THooks = void
    > extends Hooks<TMatrixHooks | H> {
        /**
         * Creates a new matrix.
         * @param nType Specifies the matrix direction.
         * @param nOrientation Specifies the matrix orientation.
         * @param nDimensions Specifies the number of columns for vertical matrices or the number of rows for horizontal matrices.
         */
        constructor(
            nType?: Types,
            nOrientation?: Orientations,
            nDimensions?: number
        );
        /** Specifies the unique id of the matrix. */
        readonly uniqueId: number;
        /** Sets the current matrix type. */
        type: Types;
        /** Sets the current matrix orientation. */
        orientation: Orientations;
        /** Sets the number of columns in the matrix. */
        columns: number;
        /** Sets the number of rows in the matrix. */
        rows: number;
        /** Retrieves the total number of items in the matrix. */
        readonly count: number;
        /** Retrieves all the items. */
        readonly all: T[];
        /** Retrieves the item at the current cursor position. */
        readonly currentItem: T | undefined;
        /** Retrieves the first item in the matrix. */
        readonly firstItem: T | undefined;
        /** Retrieves the last item in the matrix. */
        readonly lastItem: T | undefined;
        /** Retrieves the last item in the first row of the matrix. */
        readonly lastItemInFirstRow: T | undefined;
        /** Retrieves the last item in the first column of the matrix. */
        readonly lastItemInFirstColumn: T | undefined;
        /** Sets the current cursor. */
        cursor: ICursor | ISafeCursor;
        /** Retrieves the current (safe) cursor. */
        readonly safeCursor: ISafeCursor;
        /** Verifies if the cursor is at the first column. */
        readonly isCursorAtFirstColumn: boolean;
        /** Verifies if the cursor is at the last column. */
        readonly isCursorAtLastColumn: boolean;
        /** Verifies if the column cursor is past the last column. */
        readonly isCursorAtEndColumn: boolean;
        /** Verifies if the row cursor is at the first row. */
        readonly isCursorAtFirstRow: boolean;
        /** Verifies if the row cursor is at the last row. */
        readonly isCursorAtLastRow: boolean;
        /** Verifies if the row cursor is past the last row. */
        readonly isCursorAtEndRow: boolean;
        /** Verifies if the cursor is at the first item of the matrix. */
        readonly isCursorAtFirst: boolean;
        /** Verifies if the cursor is at the last item of the matrix. */
        readonly isCursorAtLast: boolean;
        /** Verifies if the cursor is past the end of the matrix. */
        readonly isCursorAtEnd: boolean;
        /** Sets the silent mode. */
        isSilent: boolean | (() => boolean);
        /**
         * Retrieves a validated cursor from the supplied cursor.
         * @param pCursor Specifies the source cursor.
         * @return Returns the validated cursor.
         */
        protected getSafeCursor(pCursor: ICursor): ISafeCursor;
        /**
         * Fired when the dimensions of the matrix are changed.
         * @param pEvent Specifies the event payload.
         * @event
         */
        protected onDimensionChange(
            pEvent: HookProps<IDimensionChangeEvent<this>>
        ): void;
        /**
         * Fired when the item count of the matrix is changed.
         * @param pEvent Specifies the event payload.
         * @event
         */
        protected onCountChange(
            pEvent: HookProps<ICountChangeEvent<this>>
        ): void;
        /**
         * Fired when the cursor position of the matrix is changed.
         * @param pEvent Specifies the event payload.
         * @event
         */
        protected onCursorChange(
            pEvent: HookProps<ICursorChangeEvent<this>>
        ): void;
        /**
         * Retrieves the index for the given column and row.
         * @param nColumn Specifies the column.
         * @param nRow Specifies the row.
         * @return Returns the item index or `-1` if the index is invalid.
         */
        indexFromColumnRow(nColumn: number, nRow: number): number;
        /**
         * Retrieves a cursor for the specified index.
         * @param nIndex Specifies the index.
         * @return Returns a cursor.
         */
        cursorFromIndex(nIndex: number | undefined): ISafeCursor;
        /**
         * Pushes an item to the matrix.
         * @param pItem Specifies the item to put in the matrix.
         * @param pOptions Specifies one or more of the following options:
         * - Index: Specifies the desired index for the item (if `undefined` the
         * supplied index in the item will be used or the item will be pushed to the
         * end of the matrix if the index is invalid);
         * - Recalculate: Specifies if the matrix needs to be recalculated;
         * - Emit: Specifies what to emit.
         * @return Returns a reference to the item.
         */
        pushToMatrix(
            pItem: T,
            pOptions?: {
                index: number | undefined;
                recalculate: boolean;
                emit: "nothing" | "push" | "change";
            }
        ): T;
        /**
         * Moves an item to the specified cursor.
         * @param pItem Specifies the item to move.
         * @param pCursor Specifies the cursor.
         * @return Returns `true` if the move succeeded.
         */
        moveTo(pItem: T, pCursor: ICursor): boolean;
        /**
         * Moves an item to the specified index position.
         * @param pItem Specifies the item to move.
         * @param nIndex Specifies the index position.
         * @return Returns `true` if the move succeeded.
         */
        moveToIndex(pItem: T, nIndex: number): boolean;
        /**
         * Moves an item to the specified cursor.
         * @param pItem Specifies the item to move.
         * @param nColumn Specifies the column.
         * @param nRow Specifies the row.
         * @return Returns `true` if the move succeeded.
         */
        moveToPosition(pItem: T, nColumn: number, nRow: number): boolean;
        /**
         * Moves an item to the specified item.
         * @param pItem Specifies the item to swap.
         * @param pTo Specifies the item to move to.
         * @param bAbove Specifies if the move position is above or below the specified item.
         * @return Returns `true` if the move succeeded.
         */
        moveToItem(pItem: T, pTo: T, bAbove: boolean): boolean;
        /**
         * Moves an item up or down.
         * @param pItem Specifies the item to move.
         * @param bUp Specifies if the move direction is up (true) or down (false).
         * @return Returns `true` if the move succeeded.
         */
        moveUpOrDown(pItem: T, bUp: boolean): boolean;
        /**
         * Swaps two items.
         * @param pA Item A.
         * @param pB Item B.
         * @return Returns `true` if the swap succeeded.
         */
        swap(pA: T, pB: T): boolean;
        /**
         * Removes an item from the matrix (or the last item if no item is specified).
         * @param pItem Specifies the item to remove.
         * @param bEmit Specifies if the events needs to be emitted.
         * @return Returns the popped item.
         */
        popFromMatrix(pItem?: T, bEmit?: boolean): T | undefined;
        /**
         * Sorts the matrix using the supplied compare functions.
         * @param fnCompare Specifies the compare function. The compare function should return a number according to the following rules:
         * - A is less than B return -1;
         * - A is greater than B return 1;
         * - A is equal to B return 0.
         *
         * To compare numbers you can simply return `nA - nB`. If no compare function is specified, the items are sorted according to the
         * original push order.
         * @param fnCompare.pA Compare item A.
         * @param fnCompare.pB Compare item B.
         * @return Returns `true` if the sort invoked an item position change.
         */
        sort(fnCompare?: (pA: T, pB: T) => number): boolean;
        /**
         * Removes all matrix items.
         * @param sMode Specifies the clear mode:
         * - `dirty`: Just removes all the items from the matrix without emitting events;
         * - `fast`: Removes the items and invokes the `onPop` method for each item;
         * - `thorough`: Removes the items and invokes the `OnPop`method and hooks for each item.
         */
        clear(sMode?: "dirty" | "fast" | "thorough"): void;
        /** Invalidates the matrix and forces the `DimensionChange` event. */
        invalidateMatrix(): void;
        /** Recalculates the matrix. */
        recalculateMatrix(): void;
        /**
         * Iterates through each item.
         * @param fnCallee Specifies the function to be invoked for each item. The item value will be exposed to the function as the first
         * argument of the argument list. Additional arguments can be specified and will be pushed to the function. If the function returns
         * `true` the iteration is stopped.
         * @param nFrom Specifies the index to start the iteration from. A negative index indicates an offset from the end of the matrix.
         * @param nTo Specifies the index to stop the iteration. A negative index indicates an offset from the end of the matrix.
         * @return Returns `true` if the iteration is stopped (break).
         */
        each(
            fnCallee: (pItem: T) => boolean | void,
            nFrom?: number,
            nTo?: number
        ): boolean;
        /**
         * Iterates through each item starting with the last item.
         * @param fnCallee Specifies the function to be invoked for each item. The item value will be exposed to the function as the first
         * argument of the argument list. Additional arguments can be specified and will be pushed to the function. If the function
         * returns `true` the iteration is stopped.
         * @param nFrom Specifies the index to start the iteration from. A negative index indicates an offset from the end of the matrix.
         * @param nTo Specifies the index to stop the iteration. A negative index indicates an offset from the end of the matrix.
         * @return Returns `true` if the iteration is be stopped (break).
         */
        reverseEach(
            fnCallee: (pItem: T) => boolean | void,
            nFrom?: number,
            nTo?: number
        ): boolean;
        /**
         * Retrieves the item at the specified index.
         * @param nIndex Specifies the index.
         * @return Returns the item or `undefined` if the item is not found.
         */
        itemAtIndex(nIndex: number): T | undefined;
        /**
         * Retrieves the item at the specified position.
         * @param nColumn Specifies the column.
         * @param nRow Specifies the row.
         * @return Returns the item or `undefined` if the item is not found.
         */
        itemAtPosition(nColumn: number, nRow: number): T | undefined;
        /**
         * Retrieves the item at the specified cursor.
         * @param pCursor Specifies the cursor.
         * @return Returns the item or `undefined` if the item is not found.
         */
        itemAtCursor(pCursor: ICursor): T | undefined;
        /**
         * Retrieves the item with the specified id.
         * @param nId Specifies the id of the item.
         * @return Returns the item or `undefined` if the item is not found.
         */
        itemFromId(nId: number): T | undefined;
        /**
         * Retrieves the last item in the specified row of the matrix.
         * @param nRow Specifies the row.
         * @return Returns a reference to the last item or `undefined` if the item is not found.
         */
        lastItemInRow(nRow: number): T | undefined;
        /**
         * Retrieves the last item in the specified column of the matrix.
         * @param nColumn Specifies the column.
         * @return Returns a reference to the last item or `undefined` if the item is not found.
         */
        lastItemInColumn(nColumn: number): T | undefined;
        /**
         * Moves the cursor to the first column.
         * @param bMoveToFirstRow Specifies if the cursor needs to be moved to the first row.
         * @return Returns the new cursor position.
         */
        moveCursorToFirstColumn(bMoveToFirstRow?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the next column.
         * @param bMoveToFirstRow Specifies if the cursor needs to be moved to the first row.
         * @return Returns the new cursor position.
         */
        moveCursorToSuccessorColumn(bMoveToFirstRow?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the previous column.
         * @param bMoveToFirstRow Specifies if the cursor needs to be moved to the first row.
         * @return Returns the new cursor position.
         */
        moveCursorToPredecessorColumn(bMoveToFirstRow?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the last column.
         * @param bMoveToFirstRow Specifies if the cursor needs to be moved to the first row.
         * @return Returns the new cursor position.
         */
        moveCursorToLastColumn(bMoveToFirstRow?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the first row.
         * @param bMoveToFirstColumn Specifies if the cursor needs to be moved to the first column.
         * @return Returns the new cursor position.
         */
        moveCursorToFirstRow(bMoveToFirstColumn?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the next row.
         * @param bMoveToFirstColumn Specifies if the cursor needs to be moved to the first column.
         * @return Returns the new cursor position.
         */
        moveCursorToSuccessorRow(bMoveToFirstColumn?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the previous row.
         * @param bMoveToFirstColumn Specifies if the cursor needs to be moved to the first column.
         * @return Returns the new cursor position.
         */
        moveCursorToPredecessorRow(bMoveToFirstColumn?: boolean): ISafeCursor;
        /**
         * Moves the row cursor to the last row.
         * @param bMoveToFirstColumn Specifies if column cursor needs to be moved to the first column.
         * @return Returns the new cursor position.
         */
        moveCursorToLastRow(bMoveToFirstColumn?: boolean): ISafeCursor;
        /**
         * Moves the cursor to the first item.
         * @return Returns the new cursor position.
         */
        moveCursorToFirst(): ISafeCursor;
        /**
         * Moves the cursor to the next item.
         * @return Returns the new cursor position.
         */
        moveCursorToSuccessor(): ISafeCursor;
        /**
         * Moves the cursor to the previous item.
         * @return Returns the new cursor position.
         */
        moveCursorToPredecessor(): ISafeCursor;
        /**
         * Moves the cursor to the last item.
         * @return Returns the new cursor position.
         */
        moveCursorToLast(): ISafeCursor;
        /**
         * Moves the cursor past the last item.
         * @return Returns the new cursor position.
         */
        moveCursorToEnd(): ISafeCursor;
    }
}

declare module "@marblecore/matrix/lib/item" {
    /** Dependencies */
    import { Enumerator } from "@marblecore/functional";
    import { HookProps, Hooks, THooks } from "@marblecore/hookup";
    import { Matrix } from "@marblecore/matrix/lib/matrix";
    import { IDispatch } from "@marblecore/matrix/lib/dispatch";
    import { ICursor } from "@marblecore/matrix/lib/cursor";
    import {
        IItemChangeEvent,
        IItemPopEvent,
        IItemPushEvent,
        TItemHooks
    } from "@marblecore/matrix/lib/events";
    import { Stack } from "@marblecore/matrix/lib/stack";
    import { Reasons } from "@marblecore/matrix/enum/reasons";
    /**
     * Implements the matrix item.
     * @param H Specifies custom hooks for the matrix (optional).
     */
    export abstract class Item<H extends THooks = void> extends Hooks<
        TItemHooks | H
    > {
        /** Contains the id enumerator. */
        static readonly enumerator: Enumerator;
        /**
         * Pushes an item to the supplied matrix.
         * @param pItem Reference to the item.
         * @param pMatrix Reference to the matrix to push the item to. If omitted
         * the current matrix of the item will be used.
         * @return Returns a reference to the item.
         */
        static pushToMatrix<T extends Item<THooks>>(
            pItem: T,
            pMatrix?: Matrix<T, THooks>
        ): T;
        /**
         * Creates a new item.
         * @param pMatrix Optional reference to the matrix.
         */
        constructor(pMatrix?: Matrix<Item<H>, THooks>);
        /** Retrieves a reference to the underlying matrix. */
        readonly parent: Matrix<Item<H>, THooks> | undefined;
        /** Specifies the unique id of the item. The id is unique within the current session. */
        readonly uniqueId: number;
        /** Overrides the automatic calculated hash value for the item. */
        hash: string;
        /** Sets the index of the item (first item is zero). */
        index: number;
        /** Returns the column index of the item in the matrix (first column is zero). */
        readonly columnIndex: number;
        /** Returns the row index of the item in the matrix (first row is zero). */
        readonly rowIndex: number;
        /** Retrieves if the item is popped (not in the matrix). */
        readonly isPopped: boolean;
        /** Specifies if the item is the first item in the column or row. */
        readonly isFirst: boolean;
        /** Specifies if the item is the last item in the column or row. */
        readonly isLast: boolean;
        /** Specifies if the item is the first item in the column. */
        readonly isFirstColumn: boolean;
        /** Specifies if the item is the last item in the column. */
        readonly isLastColumn: boolean;
        /** Specifies if the item is the first item in the row. */
        readonly isFirstRow: boolean;
        /** Specifies if the item is the last item in the row. */
        readonly isLastRow: boolean;
        /** Specifies if the item is the first item in the matrix. */
        readonly isFirstOfMatrix: boolean;
        /** Specifies if the item is the last item in the matrix. */
        readonly isLastOfMatrix: boolean;
        /** Retrieves the previous sibling of the item. */
        readonly predecessorSibling: Item<H> | undefined;
        /** Retrieves the next sibling of the item. */
        readonly successorSibling: Item<H> | undefined;
        /** Retrieves the previous item in the same column. */
        readonly predecessorInColumn: Item<H> | undefined;
        /** Retrieves the next item in the same column. */
        readonly successorInColumn: Item<H> | undefined;
        /** Retrieves the previous item in the same row. */
        readonly predecessorInRow: Item<H> | undefined;
        /** Retrieves the next item in the same row. */
        readonly successorInRow: Item<H> | undefined;
        /**
         * Retrieves a reference to the underlying matrix.
         * @return Returns a reference to the matrix.
         */
        protected matrix(): Matrix<Item<H>, THooks> | undefined;
        /**
         * Fired when the item is pushed.
         * @param pEvent Specifies the event payload.
         * @event
         */
        protected onPush(pEvent: HookProps<IItemPushEvent<Item<H>>>): void;
        /**
         * Fired when the item is changed.
         * @param pEvent Specifies the event payload.
         * @event
         */
        protected onChange(pEvent: HookProps<IItemChangeEvent<Item<H>>>): void;
        /**
         * Fired when the item is popped.
         * @param pEvent Specifies the event payload.
         * @event
         */
        protected onPop(pEvent: HookProps<IItemPopEvent<Item<H>>>): void;
        /**
         * Dispatches a matrix event.
         * @param pStack Reference to the stack.
         * @param pEvent Specifies the event to dispatch.
         */
        dispatch(
            pStack: Stack,
            pEvent:
                | IDispatch<"OnItemChange", IItemChangeEvent<Item<H>>>
                | IDispatch<"OnItemPop", IItemPopEvent<Item<H>>>
                | "Pop"
                | "Detach"
        ): void;
        /**
         * Sets the index of an item.
         * @param pStack Reference to the stack.
         * @param pMatrix Reference to the matrix.
         * @param nIndex Specifies the new index of the item.
         * @param sEmit Specifies if events needs to be emitted.
         * @param nReason Specifies the event reason.
         * @return Returns `true` if a change occured.
         */
        reindex(
            pStack: Stack,
            pMatrix: Matrix<Item<H>, THooks>,
            nIndex: number,
            sEmit: "nothing" | "auto" | "change" | "push",
            nReason: Reasons
        ): boolean;
        /**
         * Moves the item to the specified cursor.
         * @param pCursor Specifies the cursor.
         * @return Returns `true` if the move succeeded.
         */
        moveTo(pCursor: ICursor): boolean;
        /**
         * Moves the item to the specified cursor.
         * @param nIndex Specifies the index position.
         * @return Returns `true` if the move succeeded.
         */
        moveToIndex(nIndex: number): boolean;
        /**
         * Moves the item to the specified cursor.
         * @param nColumn Specifies the column.
         * @param nRow Specifies the row.
         * @return Returns `true` if the move succeeded.
         */
        moveToPosition(nColumn: number, nRow: number): boolean;
        /**
         * Moves the item to the specified item.
         * @param pTo Specifies the item to move to.
         * @param bAbove Specifies if the move position is above or below the
         * specified item.
         * @return Returns `true` if the move succeeded.
         */
        moveToItem(pTo: Item<H>, bAbove: boolean): boolean;
        /**
         * Moves the item up or down.
         * @param bUp Specifies if the move direction is up (true) or down (false).
         * @return Returns `true` if the move succeeded.
         */
        moveUpOrDown(bUp: boolean): boolean;
        /**
         * Swaps an item with another item.
         * @param pWith Specifies the item to swap with.
         * @return Returns `true` if the swap succeeded.
         */
        swap(pWith: Item<H>): boolean;
        /**
         * Removes the item from the matrix.
         * @return Returns `true` if the function succeeded.
         */
        pop(): boolean;
        /**
         * Removes the item from the matrix without emitting events.
         * @return Returns `true` if the function succeeded.
         */
        popSilent(): boolean;
        /**
         * Restores a popped item.
         * @param pMatrix Specifies a matrix to use.
         * @param nIndex Specifies the restore position.
         * @return Returns a reference to the item.
         */
        unpop(pMatrix?: Matrix<Item<H>, THooks>, nIndex?: number): boolean;
        /**
         * Restores a popped item.
         * @param pMatrix Specifies a matrix to use.
         * @param nIndex Specifies the restore position.
         * @return Returns a reference to the item.
         */
        unpopSilent(
            pMatrix?: Matrix<Item<H>, THooks>,
            nIndex?: number
        ): boolean;
    }
}

declare module "@marblecore/matrix/lib/strict/matrix" {
    import { THooks } from "@marblecore/hookup";
    import { Matrix } from "@marblecore/matrix/lib/matrix";
    import { ICursor } from "@marblecore/matrix/lib/cursor";
    import { Orientations } from "@marblecore/matrix/enum/orientations";
    import { Types } from "@marblecore/matrix/enum/types";
    import { StrictItem } from "@marblecore/matrix/lib/strict/item";
    import { IStrictItemConstructor } from "@marblecore/matrix/lib/strict/constructor";
    /**
     * Implements the strict convolution matrix. In a strict matrix an item is
     * always bound to a parent matrix and items can be created by the matrix.
     * @param P Specifies the matrix type.
     * @param T Specifies the item type.
     * @param H Specifies custom hooks for the matrix (optional).
     * @param D Specifies a data type for the item (optional).
     */
    export abstract class StrictMatrix<
        P extends StrictMatrix<P, T, H, D>,
        T extends StrictItem<P, THooks, D>,
        H extends THooks = void,
        D = undefined
    > extends Matrix<T, H> {
        /**
         * Creates a new matrix.
         * @param pItem Specifies the item constructor.
         * @param nType Specifies the matrix direction.
         * @param nOrientation Specifies the matrix orientation.
         * @param nDimensions Specifies the number of columns for vertical matrices or the number of rows for horizontal matrices.
         */
        constructor(
            pItemConstructor: IStrictItemConstructor<P, T, D>,
            nType?: Types,
            nOrientation?: Orientations,
            nDimensions?: number
        );
        /** Retrieves a reference to the item constructor. */
        readonly itemConstructor: IStrictItemConstructor<P, T, D>;
        /**
         * Appends a new item to the end of the matrix.
         * @param pData Data which is passed to the item constructor.
         * @return Returns a reference to the item.
         */
        protected appendToMatrix(pData: D): T;
        /**
         * Inserts a new item at the beginning of the matrix.
         * @param pData Data which is passed to the item constructor.
         * @return Returns a reference to the item.
         */
        protected insertToMatrix(pData: D): T;
        /**
         * Inserts a new item at the specified cursor or at the beginning of the matrix.
         * @param pData Data which is passed to the item constructor.
         * @param pCursor Specifies the cursor.
         * @return Returns a reference to the item.
         */
        protected insertToMatrixAtCursor(pData: D, pCursor: ICursor): T;
        /**
         * Inserts a new item at the specified index.
         * @param pData Data which is passed to the item constructor.
         * @param nIndex Specifies the index position.
         * @return Returns a reference to the item.
         */
        protected insertToMatrixAtIndex(pData: D, nIndex: number): T;
        /**
         * Inserts a new item at the specified index.
         * @param pData Data which is passed to the item constructor.
         * @param nColumn Specifies the column.
         * @param nRow Specifies the row.
         * @return Returns a reference to the item.
         */
        protected insertToMatrixAtPosition(
            pData: D,
            nColumn: number,
            nRow: number
        ): T;
        /**
         * Invoked when a new item instance is requested.
         * @param pData Specifies the data.
         * @return Returns the new item instance.
         * @event
         */
        onItemInstance(pData: D): T;
    }
}

declare module "@marblecore/matrix/lib/strict/item" {
    /** Dependencies */
    import { THooks } from "@marblecore/hookup";
    import { Item } from "@marblecore/matrix/lib/item";
    import { StrictMatrix } from "@marblecore/matrix/lib/strict/matrix";
    import { IDispatch } from "@marblecore/matrix/lib/dispatch";
    import {
        IItemChangeEvent,
        IItemPopEvent
    } from "@marblecore/matrix/lib/events";
    import { Stack } from "@marblecore/matrix/lib/stack";
    /**
     * Implements the strict matrix item. In a strict matrix an item is always
     * bound to a parent matrix and items can be created by the matrix.
     * @param P Specifies the matrix type.
     * @param H Specifies custom hooks for the item (optional).
     * @param D Specifies a data type for the item (optional).
     */
    export abstract class StrictItem<
        P extends StrictMatrix<P, StrictItem<P, H, D>, THooks, D>,
        H extends THooks = void,
        D = undefined
    > extends Item<H> {
        /**
         * Creates a new strict item.
         * @param pMatrix Reference to the matrix.
         */
        constructor(pMatrix: P);
        /** Retrieves a reference to the parent matrix. */
        readonly parent: P;
        /**
         * Retrieves a reference to the underlying matrix.
         * @return Returns a reference to the matrix.
         */
        protected matrix(): P;
        /**
         * Inserts a new item before this item.
         * @param pData Data which is passed to the item constructor.
         * @return Returns a reference to the item.
         */
        protected insertBefore(pData: D): StrictItem<P, H, D>;
        /**
         * Inserts a new item after this item.
         * @param pData Data which is passed to the item constructor.
         * @return Returns a reference to the item.
         */
        protected insertAfter(pData: D): StrictItem<P, H, D>;
        /**
         * Dispatches a matrix event.
         * @param pStack Reference to the stack.
         * @param pEvent Specifies the event to dispatch.
         */
        dispatch(
            pStack: Stack,
            pEvent:
                | IDispatch<
                      "OnItemChange",
                      IItemChangeEvent<StrictItem<P, H, D>>
                  >
                | IDispatch<"OnItemPop", IItemPopEvent<StrictItem<P, H, D>>>
                | "Pop"
                | "Detach"
        ): void;
    }
}

declare module "@marblecore/matrix/lib/strict/constructor" {
    /** Dependencies */
    import { THooks } from "@marblecore/hookup";
    import { StrictItem } from "@marblecore/matrix/lib/strict/item";
    import { StrictMatrix } from "@marblecore/matrix/lib/strict/matrix";
    /**
     * Describes the item constructor.
     * @param P Specifies the matrix type.
     * @param T Specifies the item type.
     * @param D Specifies a data type for the item (optional).
     */
    export interface IStrictItemConstructor<
        P extends StrictMatrix<P, T, THooks, D>,
        T extends StrictItem<P, THooks, D>,
        D
    > {
        new (pMatrix: P, pData: D): T;
    }
}

declare module "@marblecore/matrix/lib/model/model" {
    import { StrictMatrix } from "@marblecore/matrix/lib/strict/matrix";
    import { ICursor } from "@marblecore/matrix/lib/cursor";
    import { Orientations } from "@marblecore/matrix/enum/orientations";
    import { Types } from "@marblecore/matrix/enum/types";
    import { ModelItem } from "@marblecore/matrix/lib/model/item";
    /**
     * Creates a matrix model of the specified type.
     * @param T Specifies the type for the model.
     */
    export class Model<T> extends StrictMatrix<
        Model<T>,
        ModelItem<T>,
        "OnDataChange",
        T
    > {
        /**
         * Creates a new matrix.
         * @param nType Specifies the matrix direction.
         * @param nOrientation Specifies the matrix orientation.
         * @param nDimensions Specifies the number of columns for vertical matrices
         * or the number of rows for horizontal matrices.
         * @param pItems Specifies an array with initial data items.
         */
        constructor(
            nType?: Types,
            nOrientation?: Orientations,
            nDimensions?: number,
            pItems?: T[]
        );
        /**
         * Appends a new item to the end of the model.
         * @param pData Specifies the data for the item.
         * @return Returns the new item.
         */
        append(pData: T): ModelItem<T>;
        /**
         * Inserts a new item at the beginning of the model.
         * @param pData Specifies the data for the item.
         * @return Returns the new item.
         */
        insert(pData: T): ModelItem<T>;
        /**
         * Inserts a new item at the specified cursor.
         * @param pData Specifies the data for the item.
         * @param pCursor Specifies the cursor.
         * @return Returns the new item.
         */
        insertAtCursor(pData: T, pCursor: ICursor): ModelItem<T>;
        /**
         * Inserts a new item at the specified index.
         * @param pData Specifies the data for the item.
         * @param nIndex Specifies the index position.
         * @return Returns the new item.
         */
        insertAtIndex(pData: T, nIndex: number): ModelItem<T>;
        /**
         * Inserts a new item at the specified index.
         * @param pData Specifies the data for the item.
         * @param nColumn Specifies the column.
         * @param nRow Specifies the row.
         * @return Returns the new item.
         */
        insertAtPosition(pData: T, nColumn: number, nRow: number): ModelItem<T>;
    }
}

declare module "@marblecore/matrix/lib/model/item" {
    /** Dependencies */
    import { IHookPayload } from "@marblecore/hookup";
    import { StrictItem } from "@marblecore/matrix/lib/strict/item";
    import { Model } from "@marblecore/matrix/lib/model/model";
    /** Defines the `DataChange` event. */
    export interface IDataChangeEvent<T> extends IHookPayload<"OnDataChange"> {
        /** Reference to the model. */
        readonly model: Model<T>;
        /** Reference to the item. */
        readonly item: ModelItem<T>;
        /** Reference to the data. */
        readonly data: T;
    }
    /**
     * Creates a new matrix model item.
     * @param T Specifies the type for the item.
     */
    export class ModelItem<T> extends StrictItem<Model<T>, "OnDataChange", T> {
        /**
         * Creates a new item.
         * @param pData Specifies the data for the item.
         */
        constructor(pMatrix: Model<T>, pData: T);
        /** Sets the attached data. */
        data: T;
        /**
         * Inserts a new item before this item.
         * @param pData Specifies the data for the item.
         * @return Returns the new item.
         */
        insertBefore(pData: T): ModelItem<T>;
        /**
         * Inserts a new item after this item.
         * @param pData Specifies the data for the item.
         * @return Returns the new item.
         */
        insertAfter(pData: T): ModelItem<T>;
    }
}

declare module "@marblecore/matrix/lib/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Matrix } from "@marblecore/matrix/lib/matrix";
    import { Item } from "@marblecore/matrix/lib/item";
    import { ICursor } from "@marblecore/matrix/lib/cursor";
    import { Reasons } from "@marblecore/matrix/enum/reasons";
    /** Defines the matrix hooks. */
    export type TMatrixHooks =
        | "OnDimensionChange"
        | "OnCountChange"
        | "OnCursorChange"
        | "OnItemPush"
        | "OnItemChange"
        | "OnItemPop";
    /** Defines the item hooks. */
    export type TItemHooks = "OnItemPush" | "OnItemChange" | "OnItemPop";
    /** Defines the `DimensionChange` event. */
    export interface IDimensionChangeEvent<
        T extends Matrix<Item<THooks>, THooks>
    > extends IHookPayload<"OnDimensionChange"> {
        /** Reference to the matrix. */
        readonly matrix: T;
        /** Specifies the matrix volume difference since the last event. */
        readonly difference: number;
        /** Specifies the number of columns. */
        readonly columns: number;
        /** Specifies the number of rows. */
        readonly rows: number;
        /** Specifies the reason for the change. */
        readonly reason: Reasons;
    }
    /** Defines the `CountChange` event. */
    export interface ICountChangeEvent<T extends Matrix<Item<THooks>, THooks>>
        extends IHookPayload<"OnCountChange"> {
        /** Reference to the matrix. */
        readonly matrix: T;
        /** Specifies the matrix count. */
        readonly count: number;
        /** Specifies the matrix count difference since the last event. */
        readonly difference: number;
        /** Specifies the reason for the change. */
        readonly reason: Reasons;
    }
    /** Defines the `Item` event which is the base event for `IItemPushEvent` and `IItemChangeEvent`. */
    export interface IItemEvent<H extends THooks, T extends Item<THooks>>
        extends IHookPayload<H> {
        /** Reference to the current matrix. */
        readonly matrix: Matrix<T, THooks>;
        /** Reference to the previous matrix. */
        readonly matrixPrevious: Matrix<T, THooks> | undefined;
        /** Reference to the item. */
        readonly item: T;
        /** Specifies the reason for the push. */
        readonly reason: Reasons;
    }
    /** Defines the `ItemPush` event. */
    export interface IItemPushEvent<T extends Item<THooks>>
        extends IItemEvent<"OnItemPush", T> {}
    /** Defines the `ItemChange` event. */
    export interface IItemChangeEvent<T extends Item<THooks>>
        extends IItemEvent<"OnItemChange", T> {}
    /** Defines the `ItemPop` event. */
    export interface IItemPopEvent<T extends Item<THooks>>
        extends IHookPayload<"OnItemPop"> {
        /** Reference to the matrix. */
        readonly matrix: Matrix<T, THooks>;
        /** Reference to the item. */
        readonly item: T;
    }
    /** Defines the `CursorChange` event. */
    export interface ICursorChangeEvent<T extends Matrix<Item<THooks>, THooks>>
        extends IHookPayload<"OnCursorChange"> {
        /** Reference to the matrix. */
        readonly matrix: T;
        /** Contains the cursor position. */
        readonly cursor: ICursor;
    }
}

declare module "@marblecore/matrix/enum/orientations" {
    export const enum Orientations {
        /** Items a vertical oriented (columns are filled out). */
        Vertical = 0,
        /** Items are horizontal oriented (rows are filled out). */
        Horizontal = 1
    }
}

declare module "@marblecore/matrix/enum/reasons" {
    export const enum Reasons {
        /** A new item is pushed to the matrix. */
        Push = 1,
        /** An item is popped from the matrix. */
        Pop = 2,
        /** An item is moved to another position in the matrix. */
        Move = 4,
        /** An item is swapped. */
        Swap = 8,
        /** The matrix is sorted. */
        Sort = 16,
        /** A matrix recalculation occurred. */
        Recalculation = 32,
        /** An item has become the last item or the predecessor of the last item. */
        Last = 64,
        /** An item has become the first item or the successor of the first item. */
        First = 128
    }
}

declare module "@marblecore/matrix/enum/types" {
    export const enum Types {
        /** A vertical matrix expands vertical (fixed number of columns). */
        Vertical = 0,
        /** A horizontal matrix expands horizontal (fixed number of rows). */
        Horizontal = 1,
        /** A square matrix expands horizontal and vertical in a perfect square. */
        Square = 2,
        /** A horizontal square matrix expands first horizontal and then vertical. */
        SquareHorizontal = 3,
        /** A vertical square matrix expands first vertical and then horizontal. */
        SquareVertical = 4
    }
}

declare module "@marblecore/matrix/lib/dispatch" {
    /** Dependencies */
    import { HookProps, IHookPayload, THooks } from "@marblecore/hookup";
    export interface IDispatch<
        Hook extends THooks,
        T extends IHookPayload<Hook>
    > {
        hook: Hook;
        payload: HookProps<T>;
    }
}

declare module "@marblecore/matrix/lib/stack" {
    export class Stack {
        /**
         * Pushes a call to the stack.
         * @param fnCall Specifies the call.
         */
        push(fnCall: () => void): void;
        /** Flushes the stack. */
        flush(): void;
    }
}
